{$A+,B-,D+,E+,F-,G+,I+,L+,N+,P+,Q-,R-,S-,T-,V+,X+,Y+}

{DRIVER-INTERFACE FOR GRAFX2
 ---------------------------
(C) 1999,2000 by Arsne von Wyss - arsene@vonwyss.ch
Additional code by Martin Beck - spixmail@gmx.de

Rev: 20000214

This source code is freeware for non-commercial use and provided AS-IS.
You are not allowed to use or spread incomplete (inclusive this header)
copies of this code. Use at your own risk. If you do not agree with these
statements, delete this source code immediately. Please contact the
author via email for questions or bug reports.

Source code in development and subject to change frequently. No support from
the author yet.}

unit Drv_GX2;

interface

uses
	Objects,GUI,Events,Canvas,Surfaces,GrafX2,GX_Glob,MouseLib,Stream32,
				PNGLib,BMPCXLib,GIFLib,GX2Lib,CanvasFX,Drv_Gen;

type
	PScreenGrafX=^TScreenGrafX;
	TScreenGrafX=object(TScreen)
                constructor Init;
                function NewCanvas: PCanvas; virtual;
                destructor Done; virtual;
	end;
	PCanvasGrafX=^TCanvasGrafX;
	TCanvasGrafX=object(TCanvas)
                Color: TColor;
                procedure DirectPutPixel(X,Y: Integer); virtual;
                procedure DirectHLine(X1,Y,X2: Integer); virtual;
                procedure DirectVLine(X,Y1,Y2: Integer); virtual;
								procedure DirectBar(X1,Y1,X2,Y2: Integer); virtual;
                procedure DirectBlit(X1,Y1,X2,Y2,X3,Y3: Integer); virtual;
                procedure DirectWriteHLine(X1,Y,X2: Integer; const Data); virtual;
                procedure DirectReadHLine(X1,Y,X2: Integer; var Data); virtual;
                function DirectGetPixel(X,Y: Integer): Longint; virtual;
                procedure PutImage(Data: PBitmap; X,Y: Integer); virtual;
                procedure SetColor(NewColor: Longint); virtual;
                procedure Lock; virtual;
                procedure UnLock; virtual;
	end;
	PMouseDrv=^TMouseDrv;
	TMouseDrv=object(GUI.TMouse)
	private
                Hidden: Integer;
	public
                constructor Init;
                procedure Show; virtual;
                procedure Hide; virtual;
                procedure Limit(X1,Y1,X2,Y2: Integer); virtual;
                procedure Move(X,Y: Integer); virtual;
                destructor Done; virtual;
	end;
	TDriver=object(Drv_Gen.TGenericDriver)
								procedure Open; virtual;
								procedure Check; virtual;
	end;

				PLine = ^TLine;
				TLine = record
					y    : Word;
					Data : Pointer;
					Next : PLine;
				end;

				PGX2Bitmap = ^TGX2Bitmap;
				TGX2Bitmap = object (TBitmap)
					Image : TImage;
					Lines : PLine;
					constructor Init;
					destructor Done;virtual;

					procedure Reset(AWidth,AHeight: Integer; AFormat: TBitmapFormat); virtual;
					procedure Convert(AFormat: TBitmapFormat); virtual;
					function EditLine(Y: Integer): Pointer; virtual;

					procedure LoadImage(const FName : string);

					procedure CreateLineList;
					procedure FreeLineList;
				end;

var
	Driver: TDriver;
	GXScreen: TGXscreen;
	GXMouse: TMouse;

implementation

uses
	GUIUtils,NewFrontier;

var
	Screen: PScreenGrafx absolute GUI.Screen;

constructor TScreenGrafX.Init;
begin
	inherited Init(640,480);
	BitmapFormat:=bf16; {Changing the bitmap format will change the color depth}
	GXScreen.Init(Width,Height,TColorCode(BitmapFormat).Bits,mdAuto);
end;

function TScreenGrafX.NewCanvas: PCanvas;
begin
	NewCanvas:=New(PCanvasGrafX,Init);
end;

destructor TScreenGrafX.Done;
begin
	inherited Done;
	GXScreen.Close;
end;

procedure TCanvasGrafX.DirectPutPixel(X,Y: Integer);
begin
	GXScreen.PutPixel(X,Y,Color);
end;

procedure TCanvasGrafX.DirectHLine(X1,Y,X2: Integer);
begin
	GXScreen.HLine(X1,Y,Succ(X2-X1),Color);
end;

procedure TCanvasGrafX.DirectVLine(X,Y1,Y2: Integer);
begin
	GXScreen.VLine(X,Y1,Succ(Y2-Y1),Color);
end;

procedure TCanvasGrafX.DirectBar(X1,Y1,X2,Y2: Integer);
begin
	GXScreen.FillRect(X1,Y1,X2,Y2,Color,0);
end;

function TCanvasGrafX.DirectGetPixel(X,Y: Integer): Longint;
var
	RGB: TRGB;
begin
	GXScreen.DeRGB(GXScreen.GetPixel(X,Y),RGB);
	DirectGetPixel:=Longint(RGB);
end;

procedure TCanvasGrafX.DirectBlit(X1,Y1,X2,Y2,X3,Y3: Integer);
begin
	GXScreen.BitBlit(X1,Y1,X2,Y2,X3,Y3);
end;

procedure TCanvasGrafX.DirectWriteHLine(X1,Y,X2: Integer; const Data);
begin
	GXScreen.DrawLine(X1,Y,Succ(X2-X1),FarPtr(@Data),GXScreen.Mode.Bpp,drDefault);
end;

procedure TCanvasGrafX.DirectReadHLine(X1,Y,X2: Integer; var Data);
begin
	GXScreen.GetLine(X1,Y,Succ(X2-X1),FarPtr(@Data),GXScreen.Mode.Bpp);
end;

procedure TCanvasGrafX.PutImage(Data: PBitmap; X,Y: Integer);
var
	ToDraw : TSurface;
	ScrPtr  : Pointer48;
	ImgPtr  : Pointer48;
	Count   : LongInt;
	ScrPos  : Word;
	ImgPos  : LongInt;
	i,y1    : Word;
	Overlap : Word;

begin
	if not Assigned(Data) or (Data^.Format=bfNone) then Exit;
	if TypeOf(Data^)<>TypeOf(TGX2Bitmap) then
		inherited PutImage(Data,x,y)
	else
	begin
		ToDraw.Init(Data^.Width,Data^.Height,False);
		ToDraw.SurfaceCombine(@Self,-X,-Y,cmOr,False);
		if ToDraw.SurfaceCount>0 then
			if (GXScreen.Mode.Flags and mdLinear)<>0 then
				for i:=0 to ToDraw.SurfaceCount-1 do
				begin
					ScrPtr:=Ptr48(GXScreen.Mode.WritePtr,
						GXScreen.Mode.CalcOfs(X+ToDraw.SurfaceRect^[i].A.X+Offset.X,
																	Y+ToDraw.SurfaceRect^[i].A.Y+Offset.Y,
																	GXScreen.Mode)+GXScreen.Mode.DisplayOfs);
					ImgPtr:=Ptr48(PGX2Bitmap(Data)^.Image.Data,
												ToDraw.SurfaceRect^[i].A.Y*PGX2Bitmap(Data)^.Image.BPL+
												ToDraw.SurfaceRect^[i].A.X*PGX2Bitmap(Data)^.Image.Pixel+
												PGX2Bitmap(Data)^.Image.ImageStart);
					Count:=(ToDraw.SurfaceRect^[i].B.X-ToDraw.SurfaceRect^[i].A.X+1)*
								 PGX2Bitmap(Data)^.Image.Pixel;
					MoveMul2Screen(ImgPtr,ScrPtr,Count,PGX2Bitmap(Data)^.Image.BPL,
												 GXScreen.Mode.Bpl,ToDraw.SurfaceRect^[i].B.Y-
												 ToDraw.SurfaceRect^[i].A.Y+1);
				end
			else
				for i:=0 to ToDraw.SurfaceCount-1 do
				begin
					ScrPos:=GXScreen.Mode.CalcOfs(X+ToDraw.SurfaceRect^[i].A.X+Offset.X,
																				Y+ToDraw.SurfaceRect^[i].A.Y+Offset.Y,
																				GXScreen.Mode)+GXScreen.Mode.Pageadd;
					ImgPos:=ToDraw.SurfaceRect^[i].A.Y*PGX2Bitmap(Data)^.Image.BPL+
									ToDraw.SurfaceRect^[i].A.X*PGX2Bitmap(Data)^.Image.Pixel+
									PGX2Bitmap(Data)^.Image.ImageStart;
					Count:=(ToDraw.SurfaceRect^[i].B.X-ToDraw.SurfaceRect^[i].A.X+1)*
								 PGX2Bitmap(Data)^.Image.Pixel;
					for y1:=0 to (ToDraw.SurfaceRect^[i].B.Y-ToDraw.SurfaceRect^[i].A.Y) do
					begin
						if ScrPos<ScrPos+Count then
						begin
							Move2Screen(Ptr48(PGX2Bitmap(Data)^.Image.Data,ImgPos),
													Ptr48(GXScreen.Mode.WritePtr,ScrPos),Count);
							asm
								lea   si,GXScreen
								mov   ax,ds:[si].TMode.Bpl.Word
								add   ScrPos,ax
								jnc   @next
									call   IncBank
								@next:
							end;
						end
						else
						begin
							Overlap:=-ScrPos;
							Move2Screen(Ptr48(PGX2Bitmap(Data)^.Image.Data,ImgPos),
													Ptr48(GXScreen.Mode.WritePtr,ScrPos),Overlap);
							IncBank;
							Move2Screen(Ptr48(PGX2Bitmap(Data)^.Image.Data,ImgPos+Overlap),
													Ptr48(GXScreen.Mode.WritePtr,0),Count-Overlap);
							Inc(ScrPos,GXScreen.Mode.Bpl);
						end;
						Inc(ImgPos,PGX2Bitmap(Data)^.Image.Bpl);
					end;
				end;
		ToDraw.Done;
	end;
end;

procedure TCanvasGrafX.SetColor(NewColor: Longint);
var
	RGB: TRGB absolute NewColor;
begin
	inherited SetColor(NewColor);
	Color:=GXScreen.RGB(RGB.R,RGB.G,RGB.B);
end;

procedure TCanvasGrafX.Lock;
begin
	Mouse^.Hide;
end;

procedure TCanvasGrafX.UnLock;
begin
	Mouse^.Show;
end;

procedure TDriver.Open;
begin
	{Initialization: TIMER -> SCREEN -> MOUSE}
	inherited Open;
	Void(New(PScreenGrafX,Init));
	Void(New(PMouseDrv,Init));
end;

procedure TDriver.Check;
begin
	inherited Check;
	Mouse^.Update(GXMouse.Pos.X,GXMouse.Pos.Y,GXMouse.Pressed);
end;

procedure OnMouseKey(X,Y,Btn: Integer); far;
begin
	Mouse^.Update(X,Y,GXMouse.Pressed);
end;

constructor TMouseDrv.Init;
begin
	inherited Init;
	MouseLib.GlobalInitMouse(0);
	GXMouse.Init(GXScreen,0);
	GXMouse.OnMouseDown:=OnMouseKey;
	GXMouse.OnMouseUp:=OnMouseKey;
end;

procedure TMouseDrv.Show;
begin
	GXMouse.Show;
end;

procedure TMouseDrv.Hide;
begin
	GXMouse.Hide;
end;

procedure TMouseDrv.Limit(X1,Y1,X2,Y2: Integer);
begin
	GXMouse.SetWindow(X1,Y1,X2,Y2);
end;

procedure TMouseDrv.Move(X,Y: Integer);
begin
	GXMouse.SetPos(X,Y);
end;

destructor TMouseDrv.Done;
begin
	GXMouse.Done;
	MouseLib.GlobalDoneMouse;
	inherited Done;
end;

constructor TGX2Bitmap.Init;
begin
	inherited Init;
	Lines:=nil;
	Format:=bfNone;
end;

destructor TGX2Bitmap.Done;
var
	Temp,Temp2 : PLine;
begin
	Temp:=Lines;
	while Temp<>nil do
	begin
		Temp2:=Temp^.Next;
		FreeMem(Temp^.Data,Image.BPL);
		Dispose(Temp);
		Temp:=Temp2;
	end;
	if ImageOpen(Image) then CloseImage(Image);
	inherited Done;
end;

procedure TGX2Bitmap.Reset(AWidth,AHeight: Integer; AFormat: TBitmapFormat);
var
	Bpp : Word;
begin
	Width:=AWidth;
	Height:=AHeight;
	Format:=AFormat;
	ColorKey:=-1;
	if ImageOpen(Image) then CloseImage(Image);
end;

procedure TGX2Bitmap.Convert(AFormat: TBitmapFormat);
var
	TempImg : TImage;
	DestBpp : Word;
begin
	case AFormat of
		bfMono : DestBpp:=1;
		bf8    : DestBpp:=8;
		bf15   : DestBpp:=15;
		bf16   : DestBpp:=16;
		bf24   : DestBpp:=24;
		bf32   : DestBpp:=32;
	end;
	FreeLineList;
	ConvertImage(Image,TempImg,DestBpp);
	Reset(TempImg.SizeX,TempImg.SizeY,AFormat);
	LineSize:=TempImg.BPL;
	Image:=TempImg;
	CreateLineList;
end;

function TGX2Bitmap.EditLine(Y: Integer): Pointer;
var
	Seg,Ofs : LongInt;
	Temp : PLine;
begin
	Temp:=Lines;
	while (Temp<>nil)and(Temp^.y<>Y) do
		Temp:=Temp^.Next;
	if (Temp<>nil)and(Temp^.y=Y) then
	begin
		EditLine:=Temp^.Data;
	end
	else
	begin
		Ofs:=Image.ImageStart+Y*Image.BPL;
		Seg:=Image.Data+SelectorInc*(Ofs div 65536);
		EditLine:=Ptr(Word(Seg),Word(Ofs mod 65536));
	end;
end;

procedure TGX2Bitmap.LoadImage(const FName : string);
var
	FStream : TFileStream32;
	TempImg : TImage;
	AFormat : TBitmapFormat;
begin
	if not FStream.Init(FName,stOpen) then RunError;
	if LoadPNG(FStream,TempImg,32,ldAuto)<>errNoError then
	begin
		FStream.Init(FName,stOpen);
		if LoadBMP(FStream,TempImg,32,ldAuto)<>errNoError then
		begin
			FStream.Init(FName,stOpen);
			if LoadPCX(FStream,TempImg,32,ldAuto)<>errNoError then
			begin
				FStream.Init(FName,stOpen);
				if LoadGIF(FStream,TempImg,32,ldAuto)<>errNoError then
				begin
					FStream.Init(FName,stOpen);
					if LoadGX2(FStream,TempImg)<>errNoError then RunError;
				end;
			end;
		end;
	end;
	case TempImg.Bpp of
		 1 : AFormat:=bfMono;
		 8 : AFormat:=bf8;
		15 : AFormat:=bf15;
		16 : AFormat:=bf16;
		24 : AFormat:=bf24;
		32 : AFormat:=bf32;
	end;
	FreeLineList;
	Reset(TempImg.SizeX,TempImg.SizeY,AFormat);
	LineSize:=TempImg.BPL;
	Image:=TempImg;
	CreateLineList;
end;

{ diese Prozedur kreiert eine Liste, die die Linien enthlt,
	die ein 64K-Grenze passieren und dann Probleme ergeben, da
	die mit SelectorInc erhaltenen Selektoren nur eine Lnge von
	64K besitzen es also zum Runtime Error 216 kommt, wenn man darber
	kopieren will }
procedure TGX2Bitmap.CreateLineList;
var
	Offset : LongInt;
	y : Word;
	Temp : PLine;
begin
	Offset:=Image.ImageStart;
	Y:=0;
	repeat
		if Offset div 65536<>(Offset+Image.BPL)div 65536 then
		begin
			if Lines=nil then
			begin
				New(Lines);
				Temp:=Lines;
			end
			else
			begin
				Temp:=Lines;
				while Temp^.Next<>nil do
					Temp:=Temp^.Next;
				New(Temp^.Next);
				Temp:=Temp^.Next;
			end;
			Temp^.y:=y;
			GetMem(Temp^.Data,Image.BPL);
			Move2(Ptr48(Image.Data,Image.ImageStart+y*Image.BPL),
						FarPtr(Temp^.Data),Image.BPL);
			Temp^.Next:=nil;
		end;
		Inc(y);
		Inc(Offset,Image.BPL);
	until Offset=Image.ImageStart+Image.SizeY*Image.BPL;
end;

procedure TGX2Bitmap.FreeLineList;
var
	Temp,Temp2 : PLine;
begin
	Temp:=Lines;
	while Temp<>nil do
	begin
		Temp2:=Temp^.Next;
		FreeMem(Temp^.Data,Image.BPL);
		Dispose(Temp);
		Temp:=Temp2;
	end;
	Lines:=nil;
end;


begin
	Driver.Init;
end.