{$A+,B-,D+,E+,F-,G+,I+,L+,N+,P+,Q-,R-,S-,T-,V+,X+,Y+}

(* THREADS (Cooperative)
	 -------
(C) 1997,2000 by Arsne von Wyss - arsene@vonwyss.ch

Rev: 20000214

This source code is freeware for non-commercial use and provided AS-IS.
You are not allowed to use or spread incomplete (inclusive this header)
copies of this code. Use at your own risk. If you do not agree with these
statements, delete this source code immediately. Please contact the
author via email for questions or bug reports.

Source code in development and subject to change frequently. No support from
the author yet.*)

unit Threads;

interface

uses
	Objects;

type
	PThread=^TThread;
	TThread=object(TObject)
	private
		FStack: Pointer;
		FStackSize,FSP: Word;
		FNext: PThread;
		function FPrev: PThread;
		procedure FreeSelf;
	public
		Name: string;
		Terminated: Boolean;
		constructor Init(const AName: string; AStack: Word);
		destructor Done; virtual;
		procedure Execute; virtual;
		procedure Terminate;
	end;
	TMainThread=object(TThread)
		constructor Init;
		function Exists(AThread: PThread): Boolean;
		procedure Execute; virtual;
		destructor Done; virtual;
	end;

var
	MainThread: TMainThread;
	ActiveThread: PThread;

implementation

uses
	DOS;

procedure Switch; far; assembler;
asm
	push bp
	push InOutRes
	push DOSError
	les di,ActiveThread
	mov es:[di].TThread.FSP,sp
	les di,es:[di].TThread.FNext
	mov word ptr ActiveThread,di
	mov word ptr ActiveThread+2,es
	cli
	mov ss,word ptr es:[di].TThread.FStack+2
	mov sp,es:[di].TThread.FSP
	sti
	pop DOSError
	pop InOutRes
	pop bp
end;

function TThread.FPrev: PThread;
var
	Act: PThread;
begin
	Act:=FNext;
	while Act^.FNext<>@self do
		Act:=Act^.FNext;
	FPrev:=Act;
end;

constructor TThread.Init(const AName: string; AStack: Word);
begin
	inherited Init;
	Name:=AName;
	Terminated:=False;
	FStackSize:=AStack;
	GetMem(FStack,FStackSize);
	FSP:=Ofs(FStack^)+FSTackSize-2;
	FNext:=ActiveThread^.FNext;
	ActiveThread^.FNext:=@self;
	asm
		les si,self
		mov bx,es:[si]                   {VMT pointer}
		mov di,es:[si].FSP
		mov es,word ptr es:[si].FStack+2
		std
		mov ax,word ptr self+2           {SELF on stack}
		stosw
		mov ax,word ptr self
		stosw
		sub di,4                         {Dummy return address}
		mov ax,word ptr self+2           {SELF on stack}
		stosw
		mov ax,word ptr self
		stosw
		mov ax,seg TThread.Free          {Ret to TThread.Free}
		stosw
		mov ax,offset TThread.Free
		stosw
		mov ax,ds:[bx+VMTHeaderSize+4+2] {Call to TThread.Execute virtual method}
		stosw
		mov ax,ds:[bx+VMTHeaderSize+4]
		stosw
		mov es:[di],di                   {BP}
		sub ax,ax
		stosw                            {IOResult (InOutRes)}
		stosw                            {DOSError}
		mov es,word ptr self+2
		mov es:[si].FSP,di
	end;
end;

procedure TThread.FreeSelf;
begin
	FreeMem(FStack,FStackSize);
	Dispose(@self);
end;

destructor TThread.Done;
begin
	inherited Done;
	FPrev^.FNext:=FNext;           {Unhook}
	if ActiveThread=@self then asm {Selfdestroying?}
		les di,self
		mov bx,word ptr self
		mov ax,word ptr self+2
		les di,es:[di].FNext         {Change active thread...}
		mov word ptr ActiveThread,di
		mov word ptr ActiveThread+2,es
		cli
		mov ss,word ptr es:[di].TThread.FStack+2
		mov sp,es:[di].TThread.FSP
		sti
		push ax
		push bx
		call FreeSelf                {...and une the new thread to free memory}
		pop DOSError
		pop InOutRes
		pop bp
		retf                         {Restart new thread}
	end else
		FreeMem(FStack,FStackSize);
end;

procedure TThread.Execute;
begin
	Abstract;
end;

procedure TThread.Terminate;
begin
	Terminated:=True;
end;

constructor TMainThread.Init;
begin
	TObject.Init;
	FNext:=@self;
	FStack:=Ptr(SSeg,0);
	ActiveThread:=@self;
	Name:='MAIN';
end;

function TMainThread.Exists(AThread: PThread): Boolean;
var
	Act: PThread;
begin
	Act:=FNext;
	while (Act<>PThread(@self)) and (Act<>AThread) do
		Act:=Act^.FNext;
	Exists:=Act=AThread;
end;

procedure TMainThread.Execute;
begin
	while FNext<>PThread(@self) do
		Switch;
end;

destructor TMainThread.Done;
begin
	Halt;
end;

end.